# This file is part of aeb43.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"""Test AEB43
"""
import datetime as dt
import io
import os
import unittest
from decimal import Decimal

from aeb43 import AEB43

here = os.path.dirname(__file__)


class TestAEB43(unittest.TestCase):

    def setUp(self):
        self.aeb43 = AEB43(os.path.join(here, 'AEB43.txt'))

    def test_number_accounts(self):
        "Test number of accounts"
        self.assertEqual(len(self.aeb43.accounts), 1)

    def test_account(self):
        "Test account"
        account = self.aeb43.accounts[0]
        self.assertEqual(account.bank_code, '0081')
        self.assertEqual(account.branch_code, '5398')
        self.assertEqual(account.number, '0001414452')
        self.assertEqual(account.start_date, dt.date(2018, 3, 18))
        self.assertEqual(account.end_date, dt.date(2018, 3, 20))
        self.assertEqual(account.initial_balance, Decimal('3005.00'))
        self.assertEqual(account.final_balance, Decimal('2994.02'))
        self.assertEqual(account.currency, '978')
        self.assertEqual(account.name, 'DUNDER MIFFLIN')
        self.assertEqual(account.code, '   ')
        self.assertEqual(account.client_account_code, '00815398730001414452')
        self.assertEqual(account.iban, 'ES0600815398730001414452')

    def test_number_transactions(self):
        "Test number of transactions"
        account = self.aeb43.accounts[0]
        self.assertEqual(len(account.transactions), 1)

    def test_transaction(self):
        "Test transaction"
        account = self.aeb43.accounts[0]
        transaction = account.transactions[0]
        self.assertEqual(transaction.branch_code, '0901')
        self.assertEqual(transaction.transaction_date, dt.date(2018, 3, 19))
        self.assertEqual(transaction.value_date, dt.date(2018, 3, 19))
        self.assertEqual(transaction.shared_item, '12')
        self.assertEqual(transaction.own_item, '408')
        self.assertEqual(transaction.amount, Decimal('-10.98'))
        self.assertEqual(transaction.document, '0' * 10)
        self.assertEqual(transaction.reference1, '0' * 12)
        self.assertEqual(transaction.reference2, '5540014387733014')
        self.assertEqual(
            transaction.items,
            ['COMPRA TARG 5540XXXXXXXX3014 DNH*MICHA', 'EL SCOTT'])


class TestAEB43Stream(TestAEB43):

    def setUp(self):
        with io.open(os.path.join(here, 'AEB43.txt')) as f:
            self.aeb43 = AEB43(f)



# This file is part of aeb43.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import doctest
import os

here = os.path.dirname(__file__)
readme = os.path.normpath(os.path.join(here, '..', 'README.rst'))


def load_tests(loader, tests, pattern):
    if os.path.isfile(readme):
        tests.addTest(doctest.DocFileSuite(
                readme, module_relative=False,
                encoding='utf-8',
                optionflags=doctest.REPORT_ONLY_FIRST_FAILURE))
    return tests



#!/usr/bin/env python
# This file is part of aeb43.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import codecs
import os
import re

from setuptools import find_packages, setup


def read(fname):
    return codecs.open(
        os.path.join(os.path.dirname(__file__), fname), 'r', 'utf-8').read()


def get_version():
    init = read(os.path.join('aeb43', '__init__.py'))
    return re.search("__version__ = '([0-9.]*)'", init).group(1)


setup(name='aeb43',
    version=get_version(),
    description='A module to parse AEB43 files',
    long_description=read('README.rst'),
    author='Tryton',
    author_email='foundation@tryton.org',
    url='https://pypi.org/project/aeb43/',
    download_url='https://downloads.tryton.org/aeb43/',
    project_urls={
        "Bug Tracker": 'https://bugs.tryton.org/aeb43',
        "Forum": 'https://discuss.tryton.org/tags/aeb43',
        "Source Code": 'https://code.tryton.org/aeb43',
        },
    keywords='AEB43 parser',
    packages=find_packages(),
    package_data={
        'aeb43': ['AEB43.txt'],
        },
    python_requires='>=3.8',
    install_requires=['python-stdnum'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        ],
    license='BSD',
    )



# This file is part of aeb43.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"""a parser for AEB43 files
"""
__version__ = '0.1.0'
__all__ = ['AEB43']

import datetime as dt
import io
from decimal import Decimal

from stdnum.es.ccc import calc_check_digits, to_iban


def _parse_date(date):
    return dt.datetime.strptime(date, '%y%m%d').date()


class AEB43:

    def __init__(self, name, encoding=None):
        self.accounts = []

        if isinstance(name, (bytes, str)):
            with io.open(name, encoding=encoding, mode='r') as f:
                self._parse(f)
        else:
            self._parse(name)

    def _parse(self, f):
        account = transaction = None
        count = 0
        for line in self._readline(f):
            code = line[:2]
            if code == '00':
                self._parse_file_header(line)
            elif code == '11':
                count += 1
                account = Account()
                self.accounts.append(account)
                self._parse_account_header(line, account)
            elif code == '22':
                count += 1
                transaction = Transaction()
                account.transactions.append(transaction)
                self._parse_transaction(line, transaction)
            elif code == '23':
                count += 1
                self._parse_transaction_optional_item(line, transaction)
            elif code == '33':
                count += 1
                self._parse_account_end(line, account)
                account = transaction = None
            elif code == '88':
                assert count == int(line[20:26])
                account = transaction = None
                break

    @staticmethod
    def _readline(f):
        for line in f:
            yield line

    def _parse_file_header(self, line):
        self.bank_code = line[2:6]
        self.accounting_date = _parse_date(line[6:12])

    def _parse_account_header(self, line, account):
        account.bank_code = line[2:6]
        account.branch_code = line[6:10]
        account.number = line[10:20]
        account.start_date = _parse_date(line[20:26])
        account.end_date = _parse_date(line[26:32])
        sign = {
            '1': -1,
            '2': 1,
            }[line[32]]
        account.initial_balance = sign * Decimal(line[33:47]) / 100
        account.currency = line[47:50]
        account.name = line[51:77].strip()
        account.code = line[77:80]

    def _parse_transaction(self, line, transaction):
        transaction.branch_code = line[6:10]
        transaction.transaction_date = _parse_date(line[10:16])
        transaction.value_date = _parse_date(line[16:22])
        transaction.shared_item = line[22:24]
        transaction.own_item = line[24:27]
        sign = {
            '1': -1,
            '2': 1,
            }[line[27]]
        transaction.amount = sign * Decimal(line[28:42]) / 100
        transaction.document = line[42:52]
        transaction.reference1 = line[52:64]
        transaction.reference2 = line[64:80]

    def _parse_transaction_optional_item(self, line, transaction):
        transaction.items.append(line[4:42].strip())
        transaction.items.append(line[42:80].strip())

    def _parse_account_end(self, line, account):
        assert account.bank_code == line[2:6]
        assert account.branch_code == line[6:10]
        assert account.number == line[10:20]
        assert account.number_of_debit == int(line[20:25])
        assert account.total_debit == Decimal(line[25:39]) / 100
        assert account.number_of_credit == int(line[39:44])
        assert account.total_credit == Decimal(line[44:58]) / 100
        sign = {
            '1': -1,
            '2': 1,
            }[line[58]]
        account.final_balance = sign * Decimal(line[59:73]) / 100
        assert account.currency == line[73:76]

        assert (
            account.initial_balance
            - account.total_debit
            + account.total_credit) == account.final_balance


class _SlotsNone:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name in self.__slots__:
            setattr(self, name, None)


class Account(_SlotsNone):
    __slots__ = (
        'bank_code', 'branch_code', 'number', 'start_date', 'end_date',
        'initial_balance', 'final_balance', 'currency', 'name', 'code',
        'transactions')

    def __init__(self):
        super().__init__()
        self.transactions = []

    @property
    def client_account_code(self):
        check = calc_check_digits(
            self.bank_code + self.branch_code + '00' + self.number)
        return self.bank_code + self.branch_code + check + self.number

    ccc = client_account_code

    @property
    def iban(self):
        return to_iban(self.client_account_code)

    @property
    def number_of_debit(self):
        return len([t for t in self.transactions if t.amount < 0])

    @property
    def total_debit(self):
        return -sum((t.amount for t in self.transactions if t.amount < 0))

    @property
    def number_of_credit(self):
        return len([t for t in self.transactions if t.amount > 0])

    @property
    def total_credit(self):
        return -sum((t.amount for t in self.transactions if t.amount > 0))


class Transaction(_SlotsNone):
    __slots__ = (
        'branch_code', 'transaction_date', 'value_date', 'shared_item',
        'own_item', 'amount', 'document', 'reference1', 'reference2', 'items')

    def __init__(self):
        super().__init__()
        self.items = []
